#undef PACKAGE
#undef VERSION
#undef GETTEXT_PACKAGE
#undef HALT_COMMAND
#undef REBOOT_COMMAND
#undef SOUND_PROGRAM
#undef SUSPEND_COMMAND
#undef ENABLE_IPV6
#undef ENABLE_NLS
#undef ALWAYS_RESTART_SERVER
#undef HAVE_ADT
#undef HAVE_CATGETS
#undef HAVE_CHKAUTHATTR
#undef HAVE_CHPASS
#undef HAVE_CLEARENV
#undef HAVE_CRYPT
#undef HAVE_GETTEXT
#undef HAVE_LC_MESSAGES
#undef HAVE_LIBSM
#undef HAVE_LIBXDMCP
#undef HAVE_LOGINCAP
#undef HAVE_LOGINDEVPERM
#undef HAVE_LOGINRESTRICTIONS
#undef HAVE_PAM
#undef HAVE_PASSWDEXPIRED
#undef HAVE_SCHED_YIELD
#undef HAVE_SELINUX
#undef HAVE_SETENV
#undef HAVE_SETRESUID
#undef HAVE_SHADOW
#undef HAVE_SOLARIS_XINERAMA
#undef HAVE_STPCPY
#undef HAVE_SYS_SOCKIO_H
#undef HAVE_TCPWRAPPERS
#undef HAVE_UNSETENV
#undef HAVE_UT_SYSLEN
#undef HAVE_XINERAMA
#undef HAVE_XFREE_XINERAMA
#undef HAVE_XINPUT
#undef X_SERVER
#undef X_SERVER_PATH
#undef X_CONFIG_OPTIONS
#undef X_PATH
#undef XSESSION_SHELL
#undef XEVIE_OPTION
#undef HAVE_LOGIN
#undef HAVE_LOGOUT
#undef HAVE_LOGWTMP

